// Minefield.cpp : Defines the entry point for the DLL application.
//

//bzfs
#include "bzfsAPI.h"

//local
#include "pow.hpp"
#include "UMMine.h"
#include "utility.h"

//stl
#include <vector>
#include <string>
#include <cmath>
#include <algorithm>

//boost
#include <boost/array.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/format.hpp>


BZ_GET_PLUGIN_VERSION

void sendMineDeathMessage(int victimID, int killerID);

typedef boost::shared_ptr<UMMine> UMMinePtr;
typedef std::vector<UMMinePtr>::iterator ml_iterator;

/** A list of all the mines currently on the field*/
std::vector<UMMinePtr> mineList;

/** Used to keep track of the killer.  Not a good solution.*/
int lastKillerID = -1;


/** This event checks how close a player is to an existing mine, and 
    fires a world weapon shock wave if he or she is withing range.*/
class UMPlayerUpdateHandler : public bz_EventHandler {
public:
  virtual void process(bz_EventData *eventData)
  {
    if(eventData->eventType != bz_ePlayerUpdateEvent)
      return;
    
    bz_PlayerUpdateEventData *eData = dynamic_cast<bz_PlayerUpdateEventData*>(eventData);
    float dist = static_cast<float>(bz_getBZDBDouble("_shockOutRadius"));
  
    ml_iterator mine = mineList.begin();
    ml_iterator end = mineList.end();
    
    while(mine != end) {
      UMMinePtr currentMine = *mine;
      
      //Skip the player who set the mine
      if(eData->playerID == currentMine->playerID) {
        ++mine;
        continue;
      }
      
      bool withinRange = distanceLess(eData->pos, currentMine->pos.c_array(), dist);
      if(withinRange) {
        currentMine->fire();
        lastKillerID = currentMine->playerID;
        
        // Remove the mine that blew up, and terminate
        // the loop so that only one mine explodes.
        mineList.erase(mine);
        break;
      }      
      ++mine;
    }
    
  }
};


/** Used to create and set mines.*/
class UMMineCommand : public bz_CustomSlashCommandHandler {
public:
  virtual bool handle (int playerID, 
                       bzApiString /*command*/, 
                       bzApiString /*message*/, 
                       bzAPIStringList* params)
  {
    bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);
    if(!player)
      return true;
       
    bzAPIStringList &args = *params;
    
    if(args.size() == 0) {
      bz_sendTextMessage(BZ_SERVER, player->playerID, 
                         "Usage: /mine <SW | SR | CR>");
      bz_sendTextMessage(BZ_SERVER, player->playerID, 
                         "SW: fires a shockwave.");
      bz_sendTextMessage(BZ_SERVER, player->playerID, 
                         "SR: fires bullets in a spherical pattern.");
      bz_sendTextMessage(BZ_SERVER, player->playerID, 
                         "CR: fires bullets in a cylindrical pattern.");
      bz_freePlayerRecord(player);
      return true;
    }    
    
    std::string flag(player->currentFlag.c_str());
    if(flag.find("US") == std::string::npos) {
      bz_sendTextMessage(BZ_SERVER, player->playerID, 
                         "You do not have the Useless(US) flag");
      bz_freePlayerRecord(player);
      return true;
    }
    
    int i = 0;
    int size = args.size();    
    float z_offset =  static_cast<float>(bz_getBZDBDouble("_tankHeight")/2);
    
    while(i < size) {
      std::string current(args[i].c_str());
      boost::to_upper(current);
      
      if(current == "SW") {
        boost::array<float, 3> mine_pos;
        std::copy(player->pos, player->pos+3, mine_pos.begin());
        mine_pos[2] += z_offset;
        UMMinePtr mine(new ShockWaveMine(player->playerID, mine_pos));
        mineList.push_back(mine);
        bz_removePlayerFlag(player->playerID);
        break;
      }
      else if(current == "SR") {
        boost::array<float, 3> mine_pos;
        std::copy(player->pos, player->pos+3, mine_pos.begin());
        mine_pos[2] += z_offset;
        UMMinePtr mine(new SphericalRegMine(player->playerID, mine_pos));
        mineList.push_back(mine);
        bz_removePlayerFlag(player->playerID);
        break;
      }
      else if(current == "CR") {
        boost::array<float, 3> mine_pos;
        std::copy(player->pos, player->pos+3, mine_pos.begin());
        mine_pos[2] += z_offset;
        UMMinePtr mine(new CylindricalRegMine(player->playerID, mine_pos));
        mineList.push_back(mine);
        bz_removePlayerFlag(player->playerID);
        break;
      }
      ++i; 
    }
    
    bz_freePlayerRecord(player);
    
    return true;
  }
};
    

/** Used to make sure that the player's mines are removed if he or she leaves. */
class UMJoinPartHandler : public bz_EventHandler {
public:
  virtual void process(bz_EventData *eventData)
  {
    if(eventData->eventType != bz_ePlayerPartEvent)
      return;
    
    bz_PlayerJoinPartEventData *eData = dynamic_cast<bz_PlayerJoinPartEventData*>(eventData);
    
    ml_iterator begin = mineList.begin();
    ml_iterator end = mineList.end();
    
    while(begin != end) {
      UMMinePtr mine = *begin;
      if(mine->playerID == eData->playerID) {
        mineList.erase(begin);
        end = mineList.end();
      }
          
      if(begin != end)
        ++begin;
    }
  }  
};


/** Only used to send a message when a player dies.
    It tells everyone who killed him or her, rather than
    just "playerX: destroyed by the server."*/
class UMDieHandler : public bz_EventHandler {
public:
  virtual void process(bz_EventData *eventData)
  {
    if(eventData->eventType != bz_ePlayerDieEvent)
      return;
    
    bz_PlayerDieEventData *eData = dynamic_cast<bz_PlayerDieEventData*>(eventData);
    
    //Only interested in those who were killed by a mine
    if((eData->shotID < kMineShotID) || (eData->shotID > kMineShotIDMax))
      return;
    
    /* 
      DON'T USE EDATA->KILLERID, SINCE
      IT IS THE SERVER WHO IS KILLING
     */
    
    sendMineDeathMessage(eData->playerID, lastKillerID);     
  }
};


/** Used to control variables*/
class UMMineSetCommand : public bz_CustomSlashCommandHandler {
public:
  virtual bool handle (int playerID, 
                       bzApiString /*command*/, 
                       bzApiString /*message*/, 
                       bzAPIStringList* params)
  {
    bz_PlayerRecord *player = bz_getPlayerByIndex(playerID);
    if(!player)
      return true;
    
    if(!player->admin)
      return true;
  }
};

UMMineCommand mineCommand;
UMMineSetCommand mineSetComm;
UMJoinPartHandler joinPartHandler;
UMPlayerUpdateHandler playerUpdateHandler;
UMDieHandler dieHandler;

BZF_PLUGIN_CALL int bz_Load(const char* /*commandLine*/)
{
  bz_debugMessage(4,"Minefield plugin loaded");
  bz_registerCustomSlashCommand ("mine", &mineCommand);
  bz_registerCustomSlashCommand ("mine-set", &mineSetComm);
  bz_registerEvent(bz_ePlayerPartEvent, &joinPartHandler);
  bz_registerEvent(bz_ePlayerUpdateEvent, &playerUpdateHandler);
  bz_registerEvent(bz_ePlayerDieEvent, &dieHandler);
  return 0;
}

BZF_PLUGIN_CALL int bz_Unload(void)
{
  bz_debugMessage(4,"Minefield plugin unloaded");
  bz_removeCustomSlashCommand ("mine");
  bz_registerCustomSlashCommand ("mine-set", &mineSetComm);
  bz_removeEvent(bz_ePlayerPartEvent, &joinPartHandler);
  bz_removeEvent(bz_ePlayerUpdateEvent, &playerUpdateHandler);
  bz_removeEvent(bz_ePlayerDieEvent, &dieHandler);
  return 0;
}

void sendMineDeathMessage(int victimID, int killerID)
{
  bz_PlayerRecord *killer = bz_getPlayerByIndex(killerID);
  bz_PlayerRecord *victim = bz_getPlayerByIndex(victimID);
  
  if(!killer || !victim)
    return;
  
  std::string a_msg = boost::str(boost::format("%1% was destroyed by %2%'s mine") 
                                 % victim->callsign.c_str() % killer->callsign.c_str());
  std::string v_msg = boost::str(boost::format("You were destroyed by %1%'s mine") 
                                 % killer->callsign.c_str());
  
  bz_sendTextMessage(BZ_SERVER, victimID, v_msg.c_str());
  bz_sendTextMessage(BZ_SERVER, BZ_ALLUSERS, a_msg.c_str());
  
  bz_freePlayerRecord(killer);
  bz_freePlayerRecord(victim);
}



// Local Variables: ***
// mode:C++ ***
// tab-width: 8 ***
// c-basic-offset: 2 ***
// indent-tabs-mode: t ***
// End: ***
// ex: shiftwidth=2 tabstop=8
